@@ -1,4 +1,4 @@
-"""Core data classes and functions."""
+"""Scikit-rt core data classes and functions."""
 
 from collections.abc import Iterable
 from pathlib import Path
@@ -397,24 +397,40 @@ class DicomFile(Data):
     Class representing files in DICOM format, with conversion to skrt objects.
 
     **Methods:**
-    - **_Init__()** : Create instance of DicomFile class.
-    - **get_object()** : 
-    - **get_matched_attributes()** : 
-    - **set_dates_and_times()** : 
-    - **set_referenced_sop_instance_uids** :
-    - **set_slice_thickness** :
+    - **_init__()** : Create instance of DicomFile class.
+    - **get_object()** : Instantiate object of specified Scikit-rt class.
+    - **get_matched_attributes()** : Identify objects with matched attributes.
+    - **set_dates_and_times()** : Store timing for specified elements.
+    - **set_referenced_sop_instance_uids** : Store UIDs of referenced objects.
+    - **set_slice_thickness** : Store slice thickness for 3D imaging data.
     '''
 
     def __init__(self, path):
-        self.path = str(path)
+        '''
+        Create instance of DicomFile class.
+
+        **Parameter:**
+
+        path : str/pathlib.Path
+            Relative or absolute path to DICOM file.
+        '''
+        # Store absolute file path as string.
+        self.path = fullpath(path)
+
+        # Attempt to read DICOM dataset.
         try:
             self.ds = pydicom.dcmread(self.path, force=True)
         except IsADirectoryError:
             self.ds = None
+
+        # Define prefixes of dataset attributes to be used
+        # to set study and item timestamps.
         elements = {
                 "study": ["Study", "Content", "InstanceCreation"],
                 "item": ["Instance", "Content", "Series", "InstanceCreation"]
                 }
+
+        # Set object attributes.
         self.set_dates_and_times(elements)
         self.set_slice_thickness()
         self.frame_of_reference_uid = getattr(
@@ -429,13 +445,34 @@ class DicomFile(Data):
         self.set_referenced_sop_instance_uids()
 
     def get_object(self, cls, **kwargs):
+        '''
+        Instantiate object of specified Scikit-rt class.
 
+        **Parameters:**
+
+        cls : Class
+            Class from which object is to be instantiated.  In principle
+            this could be an arbitrary class, but this method is intended
+            for use with:
+            - skrt.dose.Dose;
+            - skrt.dose.Plan;
+            - skrt.image.Image;
+            - skrt.patient.Study;
+            - skrt.structures.StructureSet.
+
+        **kwrgs
+            Keyword arguments to be passed to contstructor of the class
+            from which object is to be instantiated.
+        '''
+        # Set class-specific keyword arguments.
         if cls.__name__ in ["Dose", "Plan", "StructureSet"]:
             kwargs["path"] = self.path
             kwargs["load"] = False
 
+        # Instantiate class.
         obj = cls(**kwargs)
 
+        # Set attributes specific to Study and non-Study objects.
         if cls.__name__ == "Study":
             obj.date = self.study_date
             obj.time = self.study_time
@@ -447,6 +484,8 @@ class DicomFile(Data):
             obj.time = self.item_time
             obj.timestamp = self.item_timestamp
 
+        # Set attributes for all.
+        # Some values may be null, depending on object instantiated.
         obj.modality = self.modality
         obj.frame_of_reference_uid = self.frame_of_reference_uid
         obj.referenced_image_sop_instance_uid = (
@@ -463,11 +502,28 @@ class DicomFile(Data):
         return obj
 
     def get_matched_attributes(self, others, attributes=None):
+        """
+        Identify objects with attribute values matching own attribute values.
+
+        **Parameters:**
+
+        others : list
+            List of objects to be checked for matches.
+
+        attributes : list/str, default=None
+            List of strings corresponding to attributes whose values
+            are to be checked.  If passed as a string, this is converted
+            to a single-element list.
+        """
+        # Initialise list of attributes to check.
         attributes = attributes or []
         if isinstance(attributes, str):
             attributes = [attributes]
+
+        # Initialise list of matches to all input objects.
         matches = others or []
 
+        # Loop over attributes, filtering list of matches at each iteration.
         for attribute in attributes:
             matches = [match for match in matches if
                     (getattr(match, attribute, None)
@@ -476,16 +532,33 @@ class DicomFile(Data):
         return matches 
 
     def set_dates_and_times(self, elements):
+        '''
+        Store timing information for specified elements.
 
+        **Parameter:**
+
+        elements: dict
+            Dictionary where keys are element names - for example,
+            "study", "item" - and values are DICOM dataset attributes
+            to be checked, in the order listed, for date and time information.
+        '''
+        # Define time measurements of interest, and values to be used
+        # in case of missing information.
         measurements = {"Date": "00000000", "Time": "000000"}
 
+        # Set time-measurement values for each element.
         for element, timed_items in elements.items():
             for measurement, value in measurements.items():
+                # Initialise time-measurement attribute
+                # to None (no dataset available) or fallback value.
                 attribute = f"{element}_{measurement.lower()}"
                 if self.ds is None:
                     setattr(self, attribute, None)
                     continue
                 setattr(self, attribute, value)
+
+                # Reset time-measurement attribute to value of
+                # first target attribute found in DICOM dataset.
                 for timed_item in timed_items:
                     target = f"{timed_item}{measurement}"
                     if hasattr(self.ds, target):
@@ -494,6 +567,7 @@ class DicomFile(Data):
                             setattr(self, attribute, target_value.split(".")[0])
                             break
 
+            # Set timestamp.
             if self.ds is None:
                 timestamp = None
             else:
@@ -502,6 +576,11 @@ class DicomFile(Data):
                 setattr(self, f"{element}_timestamp", f"{date}_{time}")
 
     def set_referenced_sop_instance_uids(self):
+        '''
+        Store SOP instance UIDs for referenced image, structure set, and plan.
+        '''
+
+        # Store SOP instance UID for referenced image.
         try:
             uid = self.ds.ReferencedImageSequence[0].ReferencedSOPInstanceUID
         except (IndexError, AttributeError):
@@ -516,6 +595,7 @@ class DicomFile(Data):
 
         self.referenced_image_sop_instance_uid = uid
 
+        # Store SOP instance UID for referenced structure set.
         try:
             uid = (self.ds.ReferencedStructureSetSequence[0]
                     .ReferencedSOPInstanceUID)
@@ -524,6 +604,7 @@ class DicomFile(Data):
 
         self.referenced_structure_set_sop_instance_uid = uid
 
+        # Store SOP instance UID for referenced plan.
         try:
             uid = (self.ds.ReferencedRTPlanSequence[0]
                     .ReferencedSOPInstanceUID)
@@ -533,7 +614,13 @@ class DicomFile(Data):
         self.referenced_plan_sop_instance_uid = uid
 
     def set_slice_thickness(self):
+        '''
+        Store slice thickness for 3D imaging data.
+        '''
+        # Initialise slice thickness.
         self.slice_thickness = ""
+
+        # Set slice thickness from DICOM dataset.
         if hasattr(self.ds, "SliceThickness"):
             slice_thickness = self.ds.SliceThickness
             if slice_thickness:
@@ -660,6 +747,7 @@ class Dated(PathData):
             self.timestamp = f"{self.date}_{self.time}"
 
     def get_pandas_timestamp(self):
+        """Obtain own timestamp as a pandas.Timestamp object."""
         try:
             timestamp = pd.Timestamp(''.join([self.date, self.time]))
         except ValueError:
@@ -680,8 +768,22 @@ class Dated(PathData):
         return True
 
     def get_matched_timestamps(self, others=None, delta_seconds=120):
+        """
+        Identify Dated objects that have matching timestamps, within tolerance.
+
+        **Parameters:**
+
+        others : list, default=None
+            List of objects to be checked for matching timestamps.
+
+        delta_seconds : int/float, default=120
+            Maximum time difference (seconds) between timestamps for
+            the timestamps to be considered matched.
+        """
         others = others or []
         matches = []
+
+        # Loop over input objects, checking for match with self.
         for other in others:
             try:
                 delta_time = abs(float(other.time) - float(self.time))
@@ -693,12 +795,24 @@ class Dated(PathData):
         return matches
 
     def __gt__(self, other):
+        '''
+        Define whether <self> is greater than <other>.
+
+        The comparison is based first on object date, then on time,
+        then on path.
+        '''
         for attribute in ["date", "time", "path"]:
             if getattr(self, attribute) != getattr(other, attribute):
                 return getattr(self, attribute) > getattr(other, attribute)
             return False
 
     def __ge__(self, other):
+        '''
+        Define whether <self> is greater than, or equal to, <other>.
+
+        The comparison is based first on object date, then on time,
+        then on path.
+        '''
         return (self > other) or (self == other)
 
 
@@ -821,7 +935,7 @@ def fullpath(path=""):
         expanded = os.path.realpath(tmp)
     return expanded
 
-def get_logger(name="", log_level=None):
+def get_logger(name="", log_level=None, identifier="name"):
     """
     Retrieve named event logger.
 
@@ -833,13 +947,20 @@ def get_logger(name="", log_level=None):
     log_level: string/integer/None, default=None
         Severity level for event logging.  If the value is None,
         log_level is set to the value of Defaults().log_level.
+
+    identifier: str, default="name"
+        Attribute to use to identify logger messages.  Possibilities
+        include "name", "filename", "funcName".  For a list of available
+        attributes, not all of which make sense as identifiers, see:
+        https://docs.python.org/3/library/logging.html#logrecord-attributes
     """
-    formatter = Formatter("%(name)s - %(levelname)s - %(message)s")
+    formatter = Formatter(f"%({identifier})s - %(levelname)s - %(message)s")
     handler = StreamHandler()
     handler.setFormatter(formatter)
     logger = getLogger(name)
     if not logger.handlers:
         logger.addHandler(handler)
+    log_level = log_level or Defaults().log_level
     logger.setLevel(log_level)
     return logger
 
@@ -1076,6 +1197,42 @@ def get_time_separated_objects(objs, min_delta=4, unit='hour',
 
     return time_separated_objs
 
+def get_hour_in_day(timestamp):
+    '''
+    Return a timestamp's hour in day, including fractional part.
+
+    **Parameter:**
+
+    timestamp : pandas.Timestamp
+        Timestamp for which hour in day is to be returned.
+    '''
+    if isinstance(timestamp, pd.Timestamp):
+        return timestamp.hour + timestamp.minute / 60 + timestamp.second / 3600
+
+def get_hour_in_week(timestamp):
+    '''
+    Return a timestamp's hour in week, including fractional part.
+
+    **Parameter:**
+
+    timestamp : pandas.Timestamp
+        Timestamp for which hour in week is to be returned.
+    '''
+    if isinstance(timestamp, pd.Timestamp):
+        return (24 * (timestamp.isoweekday() - 1) + get_hour_in_day(timestamp))
+
+def get_day_in_week(timestamp):
+    '''
+    Return a timestamp's day in week, including fractional part.
+
+    **Parameter:**
+
+    timestamp : pandas.Timestamp
+        Timestamp for which day in week is to be returned.
+    '''
+    if isinstance(timestamp, pd.Timestamp):
+        return get_hour_in_week(timestamp) / 24
+
 def get_interval_in_days(timestamp1, timestamp2):
     '''
     Return interval in days between two timestamps.
@@ -1146,8 +1303,10 @@ def year_fraction(timestamp):
     return year_fraction
 
 def get_uid_without_slice(uid):
+    """Obtain copy of <uid>, truncated to before final dot."""
     return ".".join(uid.split(".")[:-1])
 
+'''
 def get_sequence_value(ds=None, sequence=None, tag=None):
     value = None
     sequence_data = getattr(ds, sequence, None)
@@ -1156,15 +1315,34 @@ def get_sequence_value(ds=None, sequence=None, tag=None):
         if sequence_data:
             value = getattr(sequence_data[-1], tag, None)
     return value
+'''
 
 def get_referenced_image(referrer=None, image_types=None):
+    '''
+    Retrieve from <image_types> image object referred to by <referrer>.
+
+    **Parameters:**
+
+    referrer : object, default=None
+        Object that references an image object via its SOP instance UID.
+
+    image_types : dict, default=None
+        Dictionary where keys are imaging modalities and values are lists
+        of image objects for this modality.
+    '''
     image_types = image_types or {}
     image = None
+
+    # Search for referenced image based on matching
+    # referenced image SOP instance UID.
     for modality, images in image_types.items():
         image = get_referenced_object(referrer, images,
                 "referenced_image_sop_instance_uid", True)
         if image:
             break
+
+    # Search for referenced image based on matching
+    # frame-of-reference UID.
     if image is None:
         for modality, images in image_types.items():
             matched_attributes = DicomFile.get_matched_attributes(
@@ -1175,21 +1353,96 @@ def get_referenced_image(referrer=None, image_types=None):
     return image
 
 def get_referenced_object(referrer, others, tag, omit_slice=False):
+    '''
+    Retrieve from <others> object referred to via <tag> by <referrer>.
+
+    **Parameters:**
+
+    referrer : object
+        Object that references another object via a tag.  This tag
+        should be a reference to a SOP instance UID.
+
+    others : list
+        List of objects to be considered for identifying referenced object.
+
+    tag : str
+        String identifying object attribute that references a SOP instance UID.
+        Valid values are those defined in
+        skrt.DicomFile.set_referenced_sop_instance_uids(), namely:
+        'referenced_image_sop_instance_uid';
+        'referenced_structure_set_sop_instance_uid';
+        'referenced_plan_sop_instance_uid'.
+
+    omit_slice : bool, default=False
+        If True, disregard the last part of all UIDs, meaning the part from
+        the last dot onwards.  For imaging data, this part distinguishes
+        between different slices.
+    '''
     referenced_object = None
+
+    # Obtain SOP instance UID of referenced object.
     uid1 = getattr(referrer, tag, None)
     if uid1:
+        # Optionally omit part of UID from last dot onwards.
         if omit_slice:
             uid1 = get_uid_without_slice(uid1)
 
+        # Loop over list of objects.
         for other in others:
+            # Obtain object's SOP instance UID.
             if omit_slice:
                 uid2 = get_uid_without_slice(other.sop_instance_uid)
             else:
                 uid2 = other.sop_instance_uid
 
+            # Check for match between referenced UID and object's UID.
             if uid1 == uid2:
                 referenced_object = other
                 break
 
     return referenced_object
 
+def get_associated_image(objs, voxel_selection="most"):
+    """
+    Identify an image associated with at least one of a set of objects.
+    
+    If none of the input objects has an image attribute, None is returned.
+
+    **Parameters:**
+
+    objs : list
+        List of objects, which potentially have an image attribute,
+        identifying an associated image.
+
+    voxel_selection : str, default="most"
+        Criterion used for selecting associated image is cases where there
+        is more than one candidate.  Possible values are:
+        - None: return first image encountered;
+        - "most": return image with the highest number of voxels;
+        - "least": return image with the lowest number of voxels.
+    """
+    # Initialise variables.
+    if voxel_selection not in ["least", "most"]:
+        voxel_selection = None
+    associated_image = None
+    associated_images = {}
+
+    for obj in objs:
+        # If no selection, return first image encountered.
+        associated_image = getattr(obj, "image", None)
+        if associated_image and not voxel_selection:
+            break
+        # Store image in dictionary, with number of voxels as key.
+        if associated_image:
+            n_voxel = np.prod(associated_image.get_n_voxels())
+            associated_images[n_voxel] = associated_image
+
+    # Return image based on selection criterion.
+    if associated_images:
+        if voxel_selection == "most":
+            key = max(list(associated_images.keys()))
+        else:
+            key = min(list(associated_images.keys()))
+        associated_image = associated_images[key]
+
+    return associated_image
