@@ -12,7 +12,7 @@ import pandas as pd
 import skrt.core
 from skrt.image import Image
 from skrt.structures import StructureSet
-from skrt.dose import Dose, Plan
+from skrt.dose import Dose, Plan, remove_duplicate_doses, sum_doses
 
 # Default mode for reading patient data.
 skrt.core.Defaults({"unsorted_dicom": False})
@@ -774,6 +774,33 @@ class Patient(skrt.core.PathData):
 
     def __init__(self, path=None, exclude=None, unsorted_dicom=None,
             id_mappings=None):
+        '''
+        Create instance of Patient class.
+
+        **Parameters:**
+
+        path : str/pathlib.Path, default=None
+            Relative or absolute path to a directory containing patient data.
+            If None, the path used is the path to the current working
+            directory.
+
+        exclude : list, default=None
+            List of first-level sub-directories to disregard when
+            loading patient data organised accoding to the VoxTox model
+            (patient -> studies -> modalites).  If None, then set to
+            ["logfiles"].  Ignored if <unsorted_dicom> set to True.
+
+        unsorted_dicom : bool, default=False
+            If True, don't assume that patient data are organised
+            accoridng to the VoxTox model, and create data hierarchy
+            based on information read from DICOM files.
+
+        id_mappings : dict, default=None
+            By default, the patient identifier is set to be the name
+            of the directory that contains the patient data.  The
+            id_mappings dictionary allows mapping from the default
+            name (dictionary key) to a different name (associated value).
+        '''
 
         # Initialise parameters.
         path = path or ""
@@ -810,28 +837,68 @@ class Patient(skrt.core.PathData):
         # Add to data objects a reference to the patient object.
         self.link_study_data_to_patient()
 
+        # Initialise dose sum (outside of any study).
+        self.dose_sum = None
+        self._dose_sum_time = None
+
         # Record end time, then store initialisation time.
         toc = timeit.default_timer()
         self._init_time = (toc - tic)
 
+    def __gt__(self, other):
+        '''
+        Define whether <self> is greater than <other>.
+
+        The comparison is based on id.
+        '''
+        return self.id > other.id
+
+    def __ge__(self, other):
+        '''
+        Define whether <self> is greater than, or equal to, <other>.
+
+        The comparison is based on id.
+        '''
+        return (self > other) or (self == other)
+
     def sort_dicom(self):
+        '''
+        Create patient tree of data objects from unsorted DICOM files.
+
+        This method creates the hierarchy patient -> studies -> modalities,
+        sorting through all files in the directory pointed to be the
+        patient object's path attribute, and in all sub-directories.
+        '''
+
+        # Mapping between non-imaging DICOM modalities and
+        # terms used in Patient attributes.
         data_types = {
                 "rtplan": "plan",
                 "rtstruct": "structure_set",
                 "rtdose": "dose",
                 }
+
+        # Mapping between terms used in Patient attributes and
+        # Scikit-rt classes to which they refer.
         data_classes = {
                 "dose": Dose,
                 "image": Image,
                 "plan": Plan,
                 "structure_set": StructureSet,
                 }
+
+        # Loop through files in the directory pointed to path attribute,
+        # and in sub-directories.  Use each file to try to instantiate
+        # a skrt.core.DicomFile object.  Use instantiated objects to
+        # identify studies, then group objects by study, modality,
+        # series.
         self.studies = []
         patient_path = Path(self.path)
         file_paths = sorted(patient_path.glob("**/*"))
         for file_path in file_paths:
             dcm = skrt.core.DicomFile(file_path)
             if dcm.ds:
+                # Add to patient's list of Study objects.
                 matched_attributes = dcm.get_matched_attributes(self.studies,
                         "study_instance_uid")
                 if matched_attributes:
@@ -840,11 +907,17 @@ class Patient(skrt.core.PathData):
                     study = dcm.get_object(Study)
                     self.studies.append(study)
 
+                # Add to study objects dictionary of types for each modality.
+                # All modalities not used as keys in the data_types dictionary
+                # are mapped to modality "image".
                 dstring = data_types.get(dcm.modality, "image")
                 dtypes = getattr(study, f"{dstring}_types")
                 if not dcm.modality in dtypes:
                     dtypes[dcm.modality] = []
+
                 if "image" == dstring:
+                    # For "image" modality, group DicomFile objects
+                    # by series instance.
                     matched_attributes = dcm.get_matched_attributes(
                             dtypes[dcm.modality],
                             "series_instance_uid"
@@ -855,22 +928,36 @@ class Patient(skrt.core.PathData):
                         dcm.dicom_paths = [dcm.path]
                         dtypes[dcm.modality].append(dcm)
                 else:
+                    # For modalities other than "image",
+                    # add DicomFile objects to the relevant list
+                    # of the types dictionary.
                     dtypes[dcm.modality].append(dcm)
 
+        # First loop over studies: create Scikit-rt objects;
+        # group objects by modality.
         for study in self.studies:
             for dstring, dclass in data_classes.items():
                 dtypes = getattr(study, f"{dstring}_types")
                 for modality in dtypes:
+                    # Define datastore for each modality.
                     datastore = f"{modality}_{dstring}s"
                     setattr(study, datastore, [])
                     for idx, dcm in enumerate(dtypes[modality]):
+                        # Instantiate Scikit-rt object.
                         obj = dcm.get_object(dclass)
                         if "image" == dstring:
+                            # Transfer list of source files
+                            # from DicomFile object to Image object.
                             obj.dicom_paths = list(dcm.dicom_paths)
+                            # Subsitute Image object for DicomFile object
+                            # in dictionary of image types.
                             dtypes[modality][idx] = obj
                         getattr(study, datastore).append(obj)
 
+        # Second loop over studies: link structure sets, plans, doses.
         for study in self.studies:
+
+            # Link structure sets and plans.
             if study.plan_types and study.structure_set_types:
                 for plan in study.rtplan_plans:
                     ss = skrt.core.get_referenced_object(
@@ -879,6 +966,7 @@ class Patient(skrt.core.PathData):
                     if ss:
                         plan.set_structure_set(ss)
 
+            # Link doses and plans.
             if study.plan_types and study.dose_types:
                 for dose in study.rtdose_doses:
                     plan = skrt.core.get_referenced_object(
@@ -887,19 +975,25 @@ class Patient(skrt.core.PathData):
                     if plan:
                         dose.set_plan(plan)
 
+        # Third loop over studies: link images and non-imaging data;
+        # group non-imaging data by modality of linked image.
         for study in self.studies:
+
+            # Link images and non-imaging data.
             for dcm_modality, dstring in data_types.items():
                 dtypes = getattr(study, f"{dstring}_types")
                 dcm_datastore = f"{dcm_modality}_{dstring}s"
-                orphan_objs = []
                 for obj in getattr(study, dcm_datastore, []):
                     image = skrt.core.get_referenced_image(
                             obj, study.image_types)
                     obj.set_image(image)
 
+            # Reinforce linking of images, plans and structure sets.
             if study.plan_types and study.structure_set_types:
                 for plan in study.rtplan_plans:
                     if not plan.image:
+                        # If plan and image not directly linked,
+                        # try to link via structure set.
                         ss = skrt.core.get_referenced_object(
                                 plan, study.rtstruct_structure_sets,
                                 "referenced_structure_set_sop_instance_uid")
@@ -909,11 +1003,17 @@ class Patient(skrt.core.PathData):
 
                 for ss in study.rtstruct_structure_sets:
                     if not ss.image:
+                        # If structure set and image not directly linked,
+                        # try to link via plan.
                         for plan in ss.plans:
                             if plan.image:
                                 ss.set_image(plan.image)
                                 break
 
+            # Group non-imaging data by both their own modality and
+            # the modality of linked image.  Where there is no
+            # linked image, group non-imaging data by their own
+            # modality only.
             for dcm_modality, dstring in data_types.items():
                 dtypes = getattr(study, f"{dstring}_types")
                 dcm_datastore = f"{dcm_modality}_{dstring}s"
@@ -932,6 +1032,8 @@ class Patient(skrt.core.PathData):
                     setattr(study, dcm_datastore, orphan_objs)
                     dtypes[dcm_modality] = orphan_objs
                 else:
+                    # Discard datastores not needed
+                    # if all non-imaging data has linked image.
                     if hasattr(study, dcm_datastore):
                         delattr(study, dcm_datastore)
                     if dcm_modality in dtypes:
@@ -1004,10 +1106,11 @@ class Patient(skrt.core.PathData):
 
         return files
 
-    def combined_objs(self, attr):
+    def combined_objs(self, attr, subdir=None):
         '''
-        Get list of objects across all studies associated with
-        this patient.
+        Get list of objects across all studies associated with this patient.
+
+        Optionally restrict to studies within a given sub-directory.
 
         **Parameters:**
 
@@ -1015,34 +1118,46 @@ class Patient(skrt.core.PathData):
             Attribute name, identifying list of objects associated with a study.
             If attr ends with '_types', then objects of all listed types
             are retrieved.
+
+        subdir : str, default=None
+            Subdirectory grouping studies.  If specified, only studies in this
+            subdirectory are considered.
         '''
         all_objs = []
+        studies = self.get_subdir_studies(subdir) if subdir else self.studies
         if attr.endswith('_types'):
             dtype = attr.replace('_types', 's')
-            for study in self.studies:
+            for study in studies:
                 obj_types = getattr(study, attr, [])
                 for obj_type in sorted(obj_types):
                     all_objs.extend(getattr(study, f'{obj_type}_{dtype}'))
         else:
-            for study in self.studies:
+            for study in studies:
                 objs = getattr(study, attr, None)
                 if objs:
                     all_objs.extend(objs)
         all_objs.sort()
         return all_objs
 
-    def combined_types(self, cls):
+    def combined_types(self, cls, subdir=None):
         '''
         Get list of object types across all studies, for a given class.
 
+        Optionally restrict to studies within a given sub-directory.
+
         **Parameters:**
 
         cls : str
             Class name (case insensitive), for which types of object are
             to be retrieved.
+
+        subdir : str, default=None
+            Subdirectory grouping studies.  If specified, only studies in this
+            subdirectory are considered.
         '''
         all_types = []
-        for study in self.studies:
+        studies = self.get_subdir_studies(subdir) if subdir else self.studies
+        for study in studies:
             obj_types = getattr(study, f'{cls.lower()}_types', [])
             for obj_type in sorted(obj_types):
                 if not obj_type in all_types:
@@ -1116,7 +1231,7 @@ class Patient(skrt.core.PathData):
         return self.birth_date
 
     def get_info(self, collections=None, data_labels=None, image_types=None,
-            plan_image_type=None, treatment_image_type=None,
+            dose_types=None, plan_image_type=None, treatment_image_type=None,
             min_delta=4, unit='hour', df=False):
         '''
         Retrieve patient summary information.
@@ -1174,6 +1289,13 @@ class Patient(skrt.core.PathData):
         if image_types is None:
             image_types = self.combined_types('image')
 
+        # Ensure that list is defined for dose_types.
+        if dose_types is None:
+            dose_types = self.combined_types('dose')
+
+        # Create set combining image_types and dose_types.
+        target_types = set(image_types).union(set(dose_types))
+
         # Create dictionary for information to be returned.
         info = {}
 
@@ -1228,7 +1350,7 @@ class Patient(skrt.core.PathData):
                     # For dose data, also store information on:
                     # - maximum dose.
                     for data_type, objs in sorted(data_types.items()):
-                        if data_type not in image_types:
+                        if data_type not in target_types:
                             continue
                         file_label = f'{data_label}_{data_type}_file'
                         size_label = f'{data_label}_{data_type}_size'
@@ -1266,16 +1388,20 @@ class Patient(skrt.core.PathData):
         # Store time of image used for plan creation.
         if plan_images:
             info['plan_image_time'] = plan_images[0].get_pandas_timestamp()
+            info['plan_image_day'] = info['plan_image_time'].isoweekday()
         else:
             info['plan_image_time'] = None
+            info['plan_image_day'] = None
         info['plan_image_year'] = skrt.core.year_fraction(
                 info['plan_image_time']) 
 
         # Store time of plan creation.
         if info['plan_fraction'] is not None:
             info['plan_time'] = plan.get_pandas_timestamp()
+            info['plan_day'] = info['plan_time'].isoweekday()
         else:
             info['plan_time'] = None
+            info['plan_day'] = None
         info['plan_year'] = skrt.core.year_fraction(
                 info['plan_time']) 
 
@@ -1291,9 +1417,13 @@ class Patient(skrt.core.PathData):
         if treatment_images:
             info['treatment_start'] = treatment_images[0].get_pandas_timestamp()
             info['treatment_end'] = treatment_images[-1].get_pandas_timestamp()
+            info['treatment_start_day'] = info['treatment_start'].isoweekday()
+            info['treatment_end_day'] = info['treatment_end'].isoweekday()
         else:
             info['treatment_start'] = None
             info['treatment_end'] = None
+            info['treatment_start_day'] = None
+            info['treatment_end_day'] = None
         info['treatment_start_year'] = skrt.core.year_fraction(
                 info['treatment_start']) 
         info['treatment_end_year'] = skrt.core.year_fraction(
@@ -1388,6 +1518,8 @@ class Patient(skrt.core.PathData):
                 info['modality'] = image_type
                 info['timestamp'] = image.get_pandas_timestamp()
                 info['day'] = None
+                info['hour_in_day'] = None
+                info['hour_in_week'] = None
                 info['time_delta'] = None
                 info['days_delta'] = None
                 info['whole_days_delta'] = None
@@ -1395,6 +1527,12 @@ class Patient(skrt.core.PathData):
                 info['whole_days_total'] = None
                 if info['timestamp']:
                     info['day'] = info['timestamp'].isoweekday()
+                    info['hour_in_day'] = skrt.core.get_hour_in_day(
+                            info['timestamp'])
+                    info['hour_in_week'] = skrt.core.get_hour_in_week(
+                            info['timestamp'])
+                    info['day_in_week'] = skrt.core.get_day_in_week(
+                            info['timestamp'])
                     if time_last is not None:
                         info['time_delta'] = info['timestamp'] - time_last
                     else:
@@ -1450,6 +1588,7 @@ class Patient(skrt.core.PathData):
                 info['dose_units'] = dose.get_dose_units()
                 info['dose_type'] = dose.get_dose_type()
                 info['linked_plan'] = hasattr(dose, 'plan')
+                info['modality'] = dose_type
                 if hasattr(dose, 'plan'):
                     info['linked_plan_status'] = dose.plan.get_approval_status()
                     info['linked_plan_fraction'] = dose.plan.get_n_fraction()
@@ -1460,7 +1599,7 @@ class Patient(skrt.core.PathData):
 
         return (pd.DataFrame(all_info) if df else all_info)
 
-    def get_plan_info(self, plan_types=None, df=False):
+    def get_plan_info(self, plan_types=None, df=False, plan_filter=None):
         '''
         Retrieve information about treatment plans.
 
@@ -1474,6 +1613,13 @@ class Patient(skrt.core.PathData):
         df : bool, default=False
             If False, return summary information as a dictionary.
             If True, return summary information as a pandas dataframe.
+
+        plan_filter : str, default=None
+            String specifying if filtering is to be performed:
+            - "first_only" : retrieve information only for first plan found;
+            - "first_of_each_type" : retrieve information only for the
+              first plan found of each type;
+            - any other value : no filtering.
         '''
 
         # Ensure that list is defined for plan_types.
@@ -1491,7 +1637,7 @@ class Patient(skrt.core.PathData):
                 info['timestamp'] = plan.get_pandas_timestamp()
                 info['day'] = (info['timestamp'].isoweekday()
                         if info['timestamp'] else None)
-                info['plan_type'] = plan_type
+                info['modality'] = plan_type
                 info['plan_name'] = plan.get_name()
                 info['plan_description'] = plan.get_description()
                 info['plan_prescription_description'] = (
@@ -1519,8 +1665,93 @@ class Patient(skrt.core.PathData):
 
                 all_info.append(info)
 
+                if "first_of_each_type" == plan_filter:
+                    break
+            if "first_only" == plan_filter:
+                break
+
         return (pd.DataFrame(all_info) if df else all_info)
 
+    def get_dose_sum(self, strategies=[("FRACTION", "PLAN",), ("BEAM",)],
+            set_image=True, force=False):
+        """
+        Get dose summed over dose objects.
+
+        This method has been tested using datasets for patients from the IMPORT
+        trials.  It may not be generally valid.
+
+        **Parameters:**
+
+        strategies : list, default=[("FRACTION", "PLAN",), ("BEAM",)]
+            List of tuples of dose summation types to be used in a
+            summation strategy.  Each strategy is considered in turn,
+            until a non-zero dose sum is obtained.  The default strategy
+            if first to try summing doses with summation type "FRACTION"
+            or "PLAN", and if this gives a null result to try summing
+            doses with summation type "BEAM".
+
+        set_image : bool/skrt.image.Image, default=True
+            If True or an Image object, associate an image to the
+            summed dose, and resize the summed dose as needed to match the
+            image geometry.  The image associated is the input Image object,
+            of for boolean input is the image with most voxels associated
+            with any of the doses in the sum.
+
+        force : bool, default=False
+            If False, return the result of any previous dose summation.  If
+            True, sum doses independently of any previous result.
+        """
+
+        # If not forcing summation, return any previous result.
+        if self.dose_sum is not None and not force:
+            return self.dose_sum
+
+        # Record start time
+        tic = timeit.default_timer()
+
+        # Obtain dose objects across studies, disregarding any duplicates.
+        doses = remove_duplicate_doses(self.combined_objs("dose_types"))
+
+        # Try to sum doses using defined strategies.
+        for strategy in strategies:
+            filtered_doses = [dose for dose in doses
+                if dose.get_dose_summation_type() in strategy]
+
+            if filtered_doses:
+                self.dose_sum = sum_doses(filtered_doses)
+                break
+
+        # Define image to be associated with dose sum,
+        # and/or to be used in alternative summation strategy.
+        image = None
+        if set_image or not self.dose_sum:
+            if isinstance(set_image, Image):
+                image = set_image
+            else:
+                image = skrt.core.get_associated_image(filtered_doses)
+
+        # Try alternative summation strategy,
+        # allowing for the dose arrays to be summed having different shapes.
+        if not self.dose_sum and image:
+            doses_to_sum = []
+            for dose in filtered_doses:
+                dose_clone = dose.clone()
+                dose_clone.match_size(image)
+                doses_to_sum.append(dose_clone)
+            self.dose_sum = sum_doses(doses_to_sum)
+
+        # Associate image with summed dose,
+        # and resize summed dose as needed to match image size.
+        if set_image and self.dose_sum and image:
+            self.dose_sum.set_image(image)
+            self.dose_sum.match_size(image)
+
+        # Record end time, then store time for dose summation.
+        toc = timeit.default_timer()
+        self._dose_sum_time = (toc - tic)
+
+        return self.dose_sum
+
     def get_subdir_studies(self, subdir=""):
         """Get list of studies within a given subdirectory."""
 
